import asyncio
import logging
from enum import Enum
from json import JSONDecodeError
from typing import Dict, Text, Any, Optional

import aiohttp
from aiohttp import web, WSMsgType, WSMessage
from aiohttp.web_request import Request
from aiohttp.web_ws import WebSocketResponse

from rasax.community import utils, config
from rasax.community.api import decorators

logger = logging.getLogger(__name__)

CONNECTED_WEB_SOCKETS = {}  # type: Dict[Text, WebSocketResponse]
DEFAULT_WEBSOCKET_URL = "ws://0.0.0.0:5003/ws"


class OperationStatus(Enum):
    """Represents results of operations in websocket notifications."""

    STARTED = 0
    SUCCESS = 1
    FAILED = 2
    CANCELED = 3


class NotificationService:
    def __init__(self, websocket_url: Text = DEFAULT_WEBSOCKET_URL):
        self._websocket_url = websocket_url

        # Lazy initialization of the connections
        self._session = None  # type: Optional[aiohttp.ClientSession]
        self._websocket = None  # type: Optional[aiohttp.ClientWebSocketResponse]

    def __del__(self) -> None:
        # Close everything when the object gets destroyed.
        asyncio.ensure_future(self._close_websocket())

    async def _close_websocket(self) -> None:
        if self._websocket:
            await self._websocket.close()

        if self._session:
            await self._session.close()

    async def _connect_websocket(self) -> None:
        if not self._session:
            self._session = aiohttp.ClientSession()

        if not self._websocket:
            headers = {
                "Authorization": utils.bearer_token({"username": config.SYSTEM_USER})
            }
            self._websocket = await self._session.ws_connect(
                self._websocket_url, headers=headers
            )

    async def _send_notification(self, message: Dict[Text, Any]) -> None:
        """Send a message to the websocket server.

        Args:
            message: The message.
        """

        await self._connect_websocket()
        await self._websocket.send_json(message)

    async def send_notification_to(
        self,
        recipient_id: Optional[Text],
        topic: Text,
        status: OperationStatus,
        details: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Send a notification to a Rasa X user if the user has a connected web socket.

        Args:
            recipient_id: the id of the Rasa X user who should get the notification.
            topic: the category of the message
            status: status of the operation the notification describes.
            details: optional additional information of the notification.
        """

        if not recipient_id:
            # Handling the `None` case makes it way easier for callers of the function
            # which may or may not have the recipient
            return

        message = {
            "topic": topic,
            "status": status.name,
            "details": details,
            "recipient_id": recipient_id,
        }
        await self._send_notification(message)


def run_websocket_server(websocket_port: int = 5003) -> None:
    """Runs a websocket server."""

    app = web.Application()
    app.add_routes([web.get("/ws", _websocket_handler)])
    logger.debug("Starting web socket server on port '{}'.".format(websocket_port))

    web.run_app(app, port=websocket_port)


@decorators.insert_username_from_jwt_in_websocket_handler()
async def _websocket_handler(request: Request, username: Text) -> WebSocketResponse:
    """Handles new connections to the websocket server."""

    websocket = web.WebSocketResponse()
    await websocket.prepare(request)

    _add_websocket_connection(username, websocket)

    async for message in websocket:  # type: WSMessage
        if message.type == WSMsgType.TEXT:
            parsed_message = _get_json(message)

            if _is_from_rasa_x_server(request):
                recipient_id = parsed_message.get("recipient_id")

                if recipient_id:
                    await _forward_message_to(recipient_id, parsed_message)
                else:
                    await _forward_broadcast_to_connected_websockets(parsed_message)

    _remove_websocket_connection(websocket)
    return websocket


def _get_json(message: WSMessage) -> Dict:
    try:
        return message.json() or {}
    except JSONDecodeError:
        return {}


def _is_from_rasa_x_server(request: Request) -> bool:
    localhost_aliases = ["localhost", "127.0.0.1", "0.0.0.0", "::1"]
    return any([request.host.startswith(host) for host in localhost_aliases])


async def _forward_message_to(recipient_id: Text, message: Dict[Text, Any]) -> None:
    websocket_of_recipient = CONNECTED_WEB_SOCKETS.get(recipient_id)
    if websocket_of_recipient is not None:
        await websocket_of_recipient.send_json(message)
        logger.debug("Sent notification to recipient '{}'.".format(recipient_id))
    else:
        logger.debug(
            "Web socket connection for recipient '{}' not found.".format(recipient_id)
        )


async def _forward_broadcast_to_connected_websockets(message: Dict[Text, Any]) -> None:
    logger.debug("Broadcasting this message: {}".format(message))
    send_message_coroutines = [
        w.send_json(message) for w in CONNECTED_WEB_SOCKETS.values()
    ]
    await asyncio.gather(*send_message_coroutines)


def _add_websocket_connection(sender_id: Text, websocket: WebSocketResponse) -> None:
    CONNECTED_WEB_SOCKETS[sender_id] = websocket
    logger.debug("New web socket connection with '{}'.".format(sender_id))


def _remove_websocket_connection(websocket_to_delete: WebSocketResponse) -> None:
    matching_keys = [
        sender_id
        for sender_id, socket in CONNECTED_WEB_SOCKETS.items()
        if socket == websocket_to_delete
    ]
    for sender_id in matching_keys:
        del CONNECTED_WEB_SOCKETS[sender_id]
        logger.debug(
            "Removed websocket for '{}' from connected web sockets.".format(sender_id)
        )
