import json
import logging
from typing import Dict, Text, Any, Optional, Union

from ruamel.yaml.compat import ordereddict
from sqlalchemy import and_

from rasa.utils.io import read_yaml_file
from rasax.community import config, utils
from rasax.community.database import Environment, Project, LocalPassword
from rasax.community.database.service import DbService
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.model_service import ModelService
from rasax.community.services.stack_service import StackService, AsyncEndpointConfig
from rasax.community.services.story_service import StoryService
from rasax.community.utils import dump_yaml_to_file

logger = logging.getLogger(__name__)


def default_environments_config_local(rasa_port: Union[int, Text]) -> Dict[Text, Any]:
    stack_config = ordereddict(
        [
            (
                "production",
                ordereddict(
                    [
                        ("url", "http://localhost:{}".format(str(rasa_port))),
                        ("token", config.rasa_token),
                    ]
                ),
            ),
            (
                "worker",
                ordereddict(
                    [
                        ("url", "http://localhost:{}".format(str(rasa_port))),
                        ("token", config.rasa_token),
                    ]
                ),
            ),
        ]
    )

    return {"environments": ordereddict([("rasa", stack_config)])}


def default_stack_config():
    policies = [
        {"name": "MemoizationPolicy"},
        {"name": "KerasPolicy"},
        {"name": "MappingPolicy"},
    ]
    return ordereddict(
        [
            ("pipeline", "supervised_embeddings"),
            ("language", "en"),
            ("policies", policies),
        ]
    )


class ProjectException(Exception):
    """Exception raised for errors related to projects."""

    def __init__(self, username: Text):
        self.message = username

    def __str__(self) -> Text:
        return self.message


class SettingsService(DbService):
    def init_project(self, team: Text, project_id: Text):
        """Create project for `project_id`.

        Raise `ProjectException` if a project under that name already exists.
        """

        if self.get(team, project_id):
            raise ProjectException("Project '{}' already exists.".format(project_id))

        new_project = Project(
            project_id=project_id, team=team, config=json.dumps(default_stack_config())
        )
        self.add(new_project)

        return new_project.as_dict()

    def get(self, team: Text, project_id: Text) -> Optional[Dict[Text, Text]]:
        project = self._get_project(team, project_id)

        if project:
            return project.as_dict()

    def get_config(self, team: Text, project_id: Text) -> Optional[Dict[Text, Text]]:
        """Get model config associated with project."""

        project = self._get_project(team, project_id)

        if project:
            return project.get_model_config()

    def _get_project(self, team: Text, project_id: Text) -> Optional[Project]:
        return (
            self.query(Project)
            .filter(and_(Project.project_id == project_id, Project.team == team))
            .first()
        )

    def dump_config_in_local_mode(
        self, team: Text, project_id: Text, filename: Optional[Text] = None
    ):
        """Dump domain to `filename` in yml format."""

        if config.LOCAL_MODE:
            self._dump_config(team, project_id, filename)

    def _dump_config(
        self, team: Text, project_id: Text, filename: Optional[Text] = None
    ):
        project = self._get_project(team, project_id)
        if project:
            if not filename:
                filename = (
                    json.loads(project.config).get("path") or config.default_config_path
                )
            dump_yaml_to_file(filename=filename, content=project.get_model_config())

    def save_config(
        self,
        team: Text,
        project_id: Text,
        stack_config: Dict[Text, Any],
        config_path: Optional[Text] = None,
    ) -> Dict[Text, Any]:
        logger.debug(stack_config)

        if config_path:
            stack_config["path"] = config_path

        project = self._get_project(team, project_id)
        project.config = json.dumps(stack_config)

        return stack_config

    def inject_environments_config_from_file(self, project_id: Text, filename: Text):
        """Inject a deployment environments configuration file at `filename`
        into the db.
        """

        try:
            _config = read_yaml_file(filename)
            self.inspect_environments_config(_config)
            self.save_environments_config(project_id, _config)
            logger.debug(
                "Successfully injected deployment environments "
                "configuration from file '{}'.".format(filename)
            )
        except (ValueError, FileNotFoundError) as e:
            current_config = self.get_environments_config(project_id)
            logger.warning(
                "Could not inject deployment environments "
                "configuration from file '{}'. Details:\n{}\nYou may "
                "still use Rasa X if you currently have a "
                "working configuration. Your current configuration "
                "is:\n{}".format(filename, e, json.dumps(current_config, indent=2))
            )

    @staticmethod
    def inspect_environments_config(env_config: Dict[Text, Any]) -> None:
        """Inspect deployment environments config.

        Raise ValueError if config does not contain entries for worker or
        production services. Raise ValueError if services other than production
        or worker are added and Rasa X EE is not installed.
        """

        rasa_config = env_config.get("rasa")

        if not rasa_config:
            raise ValueError("Environment config needs to have key `rasa`.")

        if "production" not in rasa_config or "worker" not in rasa_config:
            raise ValueError(
                "Environment needs to contain entries for production "
                "and worker services."
            )

        for k, v in rasa_config.items():
            if k in ("production", "worker"):
                continue
            elif not utils.is_enterprise_installed():
                raise ValueError(
                    "Rasa X Community Edition only allows a production and "
                    "a worker environment. To use more environments, please "
                    "contact us at hi@rasa.com for an Enterprise license."
                )

    def save_environments_config(
        self, project_id: Text, environments_config: Dict[Text, Any]
    ) -> Dict[Text, Dict[Text, Dict[Text, Text]]]:
        for name, env_config in environments_config.get("rasa", {}).items():
            # delete possibly existing config
            self.delete_environment_config(name, project_id)

            env = Environment(
                name=name,
                project=project_id,
                url=env_config["url"],
                token=env_config["token"],
            )
            self.add(env)

        return self.get_environments_config(project_id)

    def delete_environment_config(self, name: Text, project_id: Text) -> None:
        to_delete = (
            self.query(Environment)
            .filter(Environment.name == name)
            .filter(Environment.project == project_id)
            .first()
        )
        if to_delete:
            self.delete(to_delete)

    def get_environments_config(self, project_id: Text) -> Dict[Text, Any]:
        envs = self.query(Environment).filter(Environment.project == project_id).all()
        envs = [e.as_dict() for e in envs]

        return {"environments": {"rasa": {list(e)[0]: e[list(e)[0]] for e in envs}}}

    def get_stack_service(
        self, environment: Text, project_id: Text = config.project_name
    ) -> Optional[StackService]:
        return self.stack_services(project_id).get(environment, None)

    def stack_services(
        self, project_id: Text = config.project_name
    ) -> Dict[Text, StackService]:
        """Create StackServices for all Stack servers."""
        environments_config = self.get_environments_config(project_id).get(
            "environments", {}
        )

        _stack_services = {}
        for k, v in environments_config.get("rasa", {}).items():
            _config = AsyncEndpointConfig(v.get("url"), token=v.get("token"))
            _stack_services[k] = StackService(
                _config,
                DataService(self.session),
                StoryService(self.session),
                DomainService(self.session),
                self,
                ModelService(config.rasa_model_dir, self.session),
            )

        return _stack_services

    def save_local_password(self, password: Text) -> None:
        """Save Rasa X password in local mode.

        Overwrite existing local password, or create a new one if none is found.
        """

        local_password = self.query(LocalPassword).first()
        if local_password:
            local_password.password = password
        else:
            local_password = LocalPassword(password=password)
            self.add(local_password)

    def get_local_password(self) -> Optional[Text]:
        """Fetch Rasa X password in local mode."""

        local_password = self.query(LocalPassword).first()
        if local_password:
            return local_password.password

        return None
