import json
import logging
import typing
from typing import Text, Optional, Dict, Any, Tuple, List, Union

import aiohttp

from rasa.core.domain import Domain
from rasa.core.trackers import DialogueStateTracker, EventVerbosity
from rasa.core.utils import concat_url, dump_obj_as_yaml_to_string
from rasa.nlu.training_data.formats import MarkdownWriter
from rasa.utils.endpoints import EndpointConfig
from rasax.community import config
from rasax.community.services.model_service import ModelService
from rasax.community.services.notification_service import OperationStatus
from rasax.community import utils
from rasax.community.constants import SHARE_YOUR_BOT_CHANNEL_NAME, DEFAULT_CHANNEL_NAME
from rasax.community.services.user_service import GUEST

if typing.TYPE_CHECKING:
    from rasax.community.services.event_service import EventService
    from rasax.community.services.model_service import ModelService
    from rasax.community.services.settings_service import SettingsService
    from rasax.community.services.data_service import DataService
    from rasax.community.services.story_service import StoryService
    from rasax.community.services.domain_service import DomainService


logger = logging.getLogger(__name__)


def format_entities(example: Dict[Text, Any]) -> List[Dict[Text, Any]]:
    ents = []
    for ent in example.get("entities", []):
        value = ent.get("value")
        if isinstance(value, dict):
            value = json.dumps(value)
        elif value is not None:
            value = str(value)
        built_entity = {
            "start": int(ent["start"]),
            "end": int(ent["end"]),
            "entity": ent["entity"],
            "value": value,
        }
        if ent.get("extractor"):
            built_entity["extractor"] = ent["extractor"]
        ents.append(built_entity)

    return ents


def nlu_format(
    data: List[Dict[Text, Any]],
    regex_features: Optional[List[Dict[Text, Any]]] = None,
    lookup_tables: Optional[List[Dict[Text, Any]]] = None,
    entity_synonyms: Optional[List[Dict[Text, Any]]] = None,
) -> Dict[Text, Dict[Text, Any]]:
    formatted_data = []
    for e in data:
        formatted_entities = format_entities(e)
        formatted_data.append(
            {"text": e["text"], "intent": e["intent"], "entities": formatted_entities}
        )

    rasa_nlu_data = {"common_examples": formatted_data}

    if regex_features:
        rasa_nlu_data["regex_features"] = regex_features
    if lookup_tables:
        rasa_nlu_data["lookup_tables"] = lookup_tables
    if entity_synonyms:
        rasa_nlu_data["entity_synonyms"] = entity_synonyms

    return {"rasa_nlu_data": rasa_nlu_data}


class AsyncEndpointConfig(EndpointConfig):
    async def request(
        self,
        method: Text = "post",
        subpath: Optional[Text] = None,
        content_type: Optional[Text] = "application/json",
        send_stack_token: bool = True,
        return_method: Text = "json",
        timeout: int = 30,
        **kwargs: Any,
    ):
        """Send a HTTP request to the endpoint.

        All additional arguments will get passed through
        to aiohttp's `session.request`."""

        # create the appropriate headers
        headers = self.headers.copy()
        if content_type:
            headers["Content-Type"] = content_type
        if "headers" in kwargs:
            headers.update(kwargs["headers"])
            del kwargs["headers"]

        # create authentication parameters
        if self.basic_auth:
            auth = aiohttp.BasicAuth(
                self.basic_auth["username"], self.basic_auth["password"]
            )
        else:
            auth = None

        url = concat_url(self.url, subpath)

        # construct GET parameters
        params = self.params.copy()

        # set the authentication token if present
        if send_stack_token and self.token:
            params[self.token_name] = self.token

        if "params" in kwargs:
            params.update(kwargs["params"])
            del kwargs["params"]

        async with aiohttp.ClientSession() as session:
            async with session.request(
                method,
                url,
                timeout=timeout,
                headers=headers,
                params=params,
                auth=auth,
                **kwargs,
            ) as resp:
                if 200 <= resp.status < 300:
                    return await getattr(resp, return_method)(), resp.status
                else:
                    return await resp.text(), resp.status


class StackService:
    """Connects to a running Rasa server.

    Used to retrieve information about models and conversations."""

    def __init__(
        self,
        async_stack_endpoint: AsyncEndpointConfig,
        data_service: "DataService",
        story_service: "StoryService",
        domain_service: "DomainService",
        settings_service: "SettingsService",
        model_service: "ModelService",
    ) -> None:

        self.async_stack_endpoint = async_stack_endpoint
        self.data_service = data_service
        self.story_services = story_service
        self.domain_service = domain_service
        self.settings_service = settings_service
        self.model_service = model_service

    async def status(self):
        """Get the status of the remote Rasa server (e.g. the version.)"""

        result, status = await self.async_stack_endpoint.request(
            subpath="/version", method="get"
        )
        if status != 200:
            return None
        else:
            return result

    async def has_active_model(self) -> bool:
        """Returns whether service has an active model."""

        result, status = await self.async_stack_endpoint.request(
            subpath="/status", method="get"
        )
        if status != 200:
            return False
        else:
            return result["fingerprint"] != {}

    async def clients(self):
        """Get a list of all conversations."""

        result, status = await self.async_stack_endpoint.request(
            subpath="/conversations", method="get"
        )
        if status != 200:
            return None
        else:
            return result

    async def initialise_tracker_for_conversation(self, conversation_id: Text) -> None:
        """Initialise a new tracker for `conversation_id`.

        This tracker will only contain an action_listen.
        """

        _ = await self.tracker_json(conversation_id)

        return None

    async def tracker(
        self,
        conversation_id: Text,
        domain: Domain,
        event_verbosity: EventVerbosity = EventVerbosity.ALL,
        until: Optional[int] = None,
    ) -> DialogueStateTracker:
        """Retrieve and recreate a tracker fetched from the remote instance."""

        tracker_json = await self.tracker_json(conversation_id, event_verbosity, until)

        tracker = DialogueStateTracker.from_dict(
            conversation_id, tracker_json.get("events", []), domain.slots
        )
        return tracker

    async def tracker_json(
        self,
        conversation_id: Text,
        event_verbosity: EventVerbosity = EventVerbosity.ALL,
        until: Optional[int] = None,
    ) -> Optional[Dict[Text, Any]]:
        """Retrieve a tracker's json representation from remote instance."""

        url = "/conversations/{}/tracker?include_events={}".format(
            conversation_id, event_verbosity.name
        )
        if until:
            url += "&until={}".format(until)

        result, status = await self.async_stack_endpoint.request(
            subpath=url, method="get"
        )

        if status != 200:
            return None
        else:
            return result

    async def append_event_to_tracker(
        self, conversation_id: Text, event: Dict[Text, Any]
    ) -> Tuple[Union[Text, Dict], int]:
        """Add an event to the tracker of a conversation."""

        return await self._modify_events(conversation_id, event)

    async def update_events(
        self,
        conversation_id: Text,
        events: List[Dict[Text, Any]],
        event_service: Optional["EventService"],
    ) -> Tuple[Union[Text, Dict], int]:
        """Update events in the tracker of a conversation."""

        if event_service:
            conversation = event_service.get_conversation_metadata_for_client(
                conversation_id
            )

            is_beginning_of_interactive_conversation = (
                conversation
                and conversation["interactive"]
                and conversation.get("n_user_messages", 0) == 0
            )

            if is_beginning_of_interactive_conversation:
                latest_event_time = conversation["latest_event_time"]

                events = event_service.modify_event_time_to_be_later_than(
                    latest_event_time, events
                )
                events = event_service.remove_leading_action_listen(events)

        return await self._modify_events(conversation_id, events, update=True)

    async def _modify_events(
        self,
        conversation_id: Text,
        event: Union[Dict[Text, Any], List[Dict[Text, Any]]],
        update: bool = False,
    ) -> Tuple[Union[Text, Dict], int]:
        """Add some more events to the tracker of a conversation."""

        url = "/conversations/{}/tracker/events".format(conversation_id)
        request_method = "put" if update else "post"

        result, status = await self.async_stack_endpoint.request(
            subpath=url, method=request_method, json=event
        )

        return result, status

    async def execute_action(
        self, conversation_id: Text, action: Dict, event_verbosity: EventVerbosity
    ) -> Tuple[Union[Text, Dict], int]:
        """Run an action in a conversation."""

        url = "/conversations/{}/execute?include_events=".format(
            conversation_id, event_verbosity.name
        )

        result, status = await self.async_stack_endpoint.request(
            subpath=url, method="post", json=action
        )

        return result, status

    async def evaluate_story(self, story: Text) -> Tuple[Union[Text, Dict], int]:
        """Evaluate a story at Core's /evaluate endpoint."""

        url = "/model/test/stories"

        result, status = await self.async_stack_endpoint.request(
            subpath=url, method="post", data=story, timeout=300
        )

        return result, status

    async def domain_json(self) -> Optional[Dict]:
        """Retrieves the domain from the stack endpoint.

        Returns:
            If the request was successful the domain as dictionary, otherwise
            `None`.
        """
        url = "/domain"
        result, status = await self.async_stack_endpoint.request(
            subpath=url, method="get", headers={"Accept": "application/json"}
        )
        if status != 200:
            return None
        else:
            return result

    async def send_message(
        self, message: Dict[Text, Text], token: Text
    ) -> Optional[List[Text]]:
        """Sends user messages to the stack Rasa webhook.

        Returns:
            If the request was successful the result as a list, otherwise
            `None`.
        """

        url = "/webhooks/rasa/webhook"

        message["input_channel"] = self._get_input_channel_from_bearer(token)

        result, status = await self.async_stack_endpoint.request(
            subpath=url,
            method="post",
            json=message,
            headers={"Authorization": token},
            send_stack_token=False,
        )

        if status != 200:
            logger.error(
                "Failed to send message to Rasa Chat webhook with "
                "status {}. Response:\n{}"
                "".format(status, result)
            )
            return None
        else:
            return result

    @staticmethod
    def _get_input_channel_from_bearer(token: Text) -> Text:
        jwt_payload = utils.verify_bearer_token(token)
        user_roles = jwt_payload.get("user", {}).get("roles", [])

        input_channel = DEFAULT_CHANNEL_NAME
        if GUEST in user_roles:
            input_channel = SHARE_YOUR_BOT_CHANNEL_NAME

        return input_channel

    async def parse(self, text: Text) -> Optional[Dict]:
        url = "/model/parse"
        result, status = await self.async_stack_endpoint.request(
            subpath=url,
            method="post",
            json={"text": text},
            headers={"Accept": "application/json"},
        )

        if status != 200:
            logger.warning(
                "Parsing the message '{}' failed with "
                "status '{}'.".format(text, status)
            )
            return None
        else:
            return result

    async def start_training_process(
        self,
        team: Text = config.team_name,
        project_id: Text = config.project_name,
        initiator: Optional[Text] = None,
    ):
        url = "/model/train"
        nlu_training_data = self.data_service.get_nlu_training_data_object()
        md_formatted_data = MarkdownWriter().dumps(nlu_training_data).strip()

        stories = self.story_services.fetch_stories(None)
        combined_stories = self.story_services.get_stories_markdown(stories)

        domain = self.domain_service.get_or_create_domain()
        domain_yaml = dump_obj_as_yaml_to_string(domain)

        _config = self.settings_service.get(team, project_id).get("config")
        config_yaml = dump_obj_as_yaml_to_string(_config)

        payload = dict(
            domain=domain_yaml,
            config=config_yaml,
            nlu=md_formatted_data,
            stories=combined_stories,
            output=config.rasa_model_dir,
            force=False,
        )

        from rasax.community.services.notification_service import NotificationService

        notification_service = NotificationService()
        await notification_service.send_notification_to(
            initiator, topic="TRAINING", status=OperationStatus.STARTED
        )

        result, status = await self.async_stack_endpoint.request(
            subpath=url,
            method="post",
            json=payload,
            return_method="read",
            timeout=24 * 60 * 60,  # 24 hours
        )

        if status == 200:
            model_name = await self._save_model_to_disk(result, project_id)
            await notification_service.send_notification_to(
                initiator,
                topic="TRAINING",
                status=OperationStatus.SUCCESS,
                details={"model": model_name},
            )
        else:
            logger.warning(
                "Model training failed with status '{}'  and message '{}'."
                "".format(status, result)
            )

            await notification_service.send_notification_to(
                initiator,
                topic="TRAINING",
                status=OperationStatus.FAILED,
                details={"status": status, "message": result},
            )

    async def _save_model_to_disk(self, result: bytes, project_id: Text) -> Text:
        from rasa.cli import utils
        from rasax.community.utils import write_bytes_to_file
        import os

        # save new model in model service
        model_path = utils.create_output_path(config.rasa_model_dir)
        write_bytes_to_file(model_path, result)

        model_name = os.path.basename(model_path).split(".tar.gz")[0]
        await self.model_service.save(project_id, model_name, model_path)

        return model_name

    async def evaluate_intents(
        self, training_data: List[Dict[Text, Any]], model_path: Text
    ) -> Tuple[Optional[Dict], int]:
        url = "/model/test/intents?model={}".format(model_path)

        result, status = await self.async_stack_endpoint.request(
            subpath=url, method="post", json=nlu_format(training_data)
        )

        if status != 200:
            logger.warning(
                "Evaluating the intents failed with status '{}' "
                "and message '{}'.".format(status, result)
            )
            return None, status
        else:
            return result, status

    async def predict_next_action(
        self, conversation_id: Text, included_events: Text = "ALL"
    ) -> Tuple[Optional[Dict], int]:

        url = "/conversations/{}/predict?include_events={}".format(
            conversation_id, included_events
        )

        result, status = await self.async_stack_endpoint.request(
            subpath=url, method="post"
        )
        return result, status
