import asyncio
import os

import aiohttp
import pytest

import rasax.community.utils as rasax_utils
import rasax.community.services.notification_service as notifications
from rasax.community import utils, config

os.environ["JWT_SECRET"] = config.jwt_secret

TEST_WEBSOCKET_PORT = 5003
TEST_WEBSOCKET_URL = "ws://127.0.0.1:{}/ws".format(TEST_WEBSOCKET_PORT)


def run_websocket():
    notifications.run_websocket_server(TEST_WEBSOCKET_PORT)


@pytest.fixture
async def websocket_process():
    websocket_server_process = rasax_utils.run_in_process(run_websocket, is_local=True)
    # Wait a bit until websocket server is running
    await asyncio.sleep(2)
    yield websocket_process

    # Stop websocket server
    websocket_server_process.terminate()


async def test_failing_authentication(websocket_process):
    message_received = False

    async with aiohttp.ClientSession() as session:
        async with session.ws_connect(TEST_WEBSOCKET_URL) as ws:
            async for _ in ws:
                # This block should not be executed since connection is closed before
                message_received = True
                # break out of the loop
                break

    assert ws.closed
    assert not message_received


async def test_send_notification(websocket_process):
    notifier = notifications.NotificationService(TEST_WEBSOCKET_URL)
    messages_received = False

    connected_test_user = "test_send_notification"
    authentication_header = {
        "Authorization": utils.bearer_token({"username": connected_test_user})
    }
    # Check if we get a status update from the socket
    async with aiohttp.ClientSession() as session:
        async with session.ws_connect(
            TEST_WEBSOCKET_URL, headers=authentication_header
        ) as ws:
            test_topic = "testing"
            test_status = notifications.OperationStatus.SUCCESS

            await notifier.send_notification_to(
                connected_test_user, topic=test_topic, status=test_status
            )

            async for message in ws:
                assert message.type == aiohttp.WSMsgType.TEXT
                parsed = message.json()
                assert parsed["topic"] == test_topic
                assert parsed["status"] == test_status.name
                await ws.close()

                messages_received = True
                # break out of the loop
                break

    assert messages_received


async def test_send_notification_to_somebody_else(websocket_process):
    notifier = notifications.NotificationService(TEST_WEBSOCKET_URL)

    connected_test_user = "test_send_notification"
    authentication_header = {
        "Authorization": utils.bearer_token({"username": connected_test_user})
    }
    # Check if we get a status update from the socket
    async with aiohttp.ClientSession() as session:
        async with session.ws_connect(
            TEST_WEBSOCKET_URL, headers=authentication_header
        ) as ws:
            test_topic = "testing"
            test_status = notifications.OperationStatus.SUCCESS

            await notifier.send_notification_to(
                "some other user",
                topic=test_topic,
                status=notifications.OperationStatus.FAILED,
            )

            await notifier.send_notification_to(
                connected_test_user, topic=test_topic, status=test_status
            )

            async for message in ws:
                # Only a message with topic `SUCCESS` should be received
                assert message.json()["status"] == test_status.name
                # break out of the loop
                break


def test_add_websocket_connection():
    test_users = ["user1", "user2"]
    # Use index as simple replacement for the actual websocket connection object
    for dummy_object, user in enumerate(test_users):
        notifications._add_websocket_connection(user, dummy_object)

    assert all([user in notifications.CONNECTED_WEB_SOCKETS for user in test_users])


def test_remove_websocket_connection():
    # Insert users
    test_users = ["user1", "user2"]
    # Use index as simple replacement for the actual websocket connection object
    for dummy_object, user in enumerate(test_users):
        notifications._add_websocket_connection(user, dummy_object)

    notifications._remove_websocket_connection(1)

    assert test_users[1] not in notifications.CONNECTED_WEB_SOCKETS
    assert len(notifications.CONNECTED_WEB_SOCKETS) == 1
