import pytest
import sqlalchemy.dialects
import sqlalchemy.exc
from sqlalchemy.orm import Session

import rasax.community.database.utils as db_utils
from rasax.community.database import Role
from rasax.community.utils import (
    query_result_to_dict,
    bearer_token,
    verify_bearer_token,
    encode_jwt_token,
)
from tests.unit import conftest
from tests.unit.conftest import insert_test_user


# Overwrite fixture to run per function
@pytest.fixture()
def session():
    return next(conftest.session())


def test_query_result_to_dict():
    query_result = ["zero", "one", "two", "three"]
    fields_query = [("d", True), ("a.c", True), ("a.b", True), ("b.c.d", True)]

    actual = query_result_to_dict(query_result, fields_query)

    assert actual["d"] == query_result[0]
    assert actual["a"]["c"] == query_result[1]
    assert actual["a"]["b"] == query_result[2]
    assert actual["b"]["c"]["d"] == query_result[3]


def test_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, True)
    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


def test_turn_off_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, False)
    if isinstance(db_connection.dialect, sqlalchemy.dialects.sqlite.dialect):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


def test_turn_off_then_on_foreign_key_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, False)
    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, True)

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


def test_verify_bearer_token():
    payload = {"test": "test"}
    encoded = bearer_token(payload)

    assert verify_bearer_token(encoded) == payload


@pytest.mark.parametrize(
    "token",
    [
        bearer_token({}, jwt_secret="wrong secret"),
        encode_jwt_token({}),
        "Other prefix " + encode_jwt_token({}),
    ],
)
def test_invalid_bearer_tokens(token):
    with pytest.raises(ValueError):
        verify_bearer_token(token)
